package org.supermmx.ginger;

import org.supermmx.peony.net.PeonyChannel;
import org.supermmx.peony.net.PeonyChannelCallback;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory; 

import java.io.OutputStream;
import java.io.FileOutputStream;
import java.io.BufferedOutputStream;
import java.nio.ByteBuffer;

/**
 * Telnet protocol parser.
 *
 * @author SuperMMX
 */
public class GingerTelnet implements PeonyChannelCallback {
    /*===================*
     *  Telnet Commands  *
     *===================*/
    /**
     * Interpret As Command
     */
    public static final byte TC_IAC  = (byte) 255; // 0xff
    /**
     * DONT
     */
    public static final byte TC_DONT = (byte) 254; // 0xfe
    /**
     * DO
     */
    public static final byte TC_DO   = (byte) 253; // 0xfd
    /**
     * WONT
     */
    public static final byte TC_WONT = (byte) 252; // 0xfc
    /**
     * WILL
     */
    public static final byte TC_WILL = (byte) 251; // 0xfb
    /**
     * Subnegociation Begin
     */
    public static final byte TC_SB   = (byte) 250; // 0xfa
    /**
     * Go Ahead
     */
    public static final byte TC_GA   = (byte) 249; // 0xf9
    /**
     * Erase Line
     */
    public static final byte TC_EL   = (byte) 248; // 0xf8
    /**
     * Erase Character
     */
    public static final byte TC_EC   = (byte) 247; // 0xf7
    /**
     * Are You There
     */
    public static final byte TC_AYT  = (byte) 246; // 0xf6
    /**
     * Abort Output
     */
    public static final byte TC_AO   = (byte) 245; // 0xf5
    /**
     * Interrupt Process
     */
    public static final byte TC_IP   = (byte) 244; // 0xf4
    /**
     * Break
     */
    public static final byte TC_BRK  = (byte) 243; // 0xf3
    /**
     * Data Mark
     */
    public static final byte TC_DM   = (byte) 242; // 0xf2
    /**
     * NO oPeration
     */
    public static final byte TC_NOP  = (byte) 241; // 0xf1
    /**
     * Subnegociation End
     */
    public static final byte TC_SE   = (byte) 240; // 0xf0

    /*==================*
     *  Telnet Options  *
     *==================*/
    /**
     * Transmit Binary
     */
    public static final byte TO_BINARY = (byte) 0;
    /**
     * Echo
     */
    public static final byte TO_ECHO = (byte) 1;
    /**
     * Supress Go Ahead
     */
    public static final byte TO_SUPPRESS_GO_AHEAD = (byte) 3;
    /**
     * Terminal Type
     */
    public static final byte TO_TERMINAL_TYPE = (byte) 24;
    /**
     * Negotiate About Window Size.
     */
    public static final byte TO_NAWS = (byte) 31;

    /**
     * Client sends back the terminal type
     */
    public static final byte TERMINAL_TYPE_IS = 0;
    /**
     * Server requests the terminal type
     */
    public static final byte TERMINAL_TYPE_SEND = 1;

    /**
     * Telnet State.
     * 
     * INIT --(IAC)--> IAC --(W/WT/D/DT)--> CMD --(XXX)--> INIT
     *  |               |
     *  |               |--(SB)--> SUB --(XXX)--> SUB_OPT
     *  |                                               |
     *  ----------<------------(SE)-- SUB_IAC <--(IAC)--|
     */
    private static enum TelnetState {
        INIT, IAC, CMD, SUB,
        SUB_OPT, SUB_IAC
    }

    private static Log log = LogFactory.getLog(GingerTelnet.class);

    private GingerSession session;

    /**
     * Output stream for data
     */
    private OutputStream os;

    /**
     * Telnet state
     */
    private TelnetState state;
    /**
     * Telnet data
     */
    private TelnetData data;

    /**
     * Buffer passed to upstream for further process 
     */
    private ByteBuffer outBuf;

    public GingerTelnet(GingerSession session) {
        this.session = session;

        state = TelnetState.INIT;
        data = new TelnetData();
    }

    /*========================*
     *  PeonyChannelCallback  *
     *========================*/
    public void connected(PeonyChannel channel) {
        try {
            log.debug("Connected");
            if (os != null) {
                os.flush();
                os.close();
            }
            // FIXME: with dest name
            os = new BufferedOutputStream(new FileOutputStream("rawdata.log"));
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public void disconnected(PeonyChannel channel) {
        log.debug("Disconnected");
        if (os != null) {
            try {
                os.flush();
                os.close();
            } catch (Exception e) {
                e.printStackTrace();
            }
            os = null;
        }
    }

    /**
     * Process incoming data.
     *
     * @param channel the underlying channel
     * @param buf the byte buffer
     */
    public void process(PeonyChannel channel, byte[] buf) {
        log.debug("Data received, size = " + buf.length);
        if (os != null) {
            try {
                os.write(buf);
                os.write("\n\n".getBytes());
                os.flush();
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        ByteBuffer ib = ByteBuffer.wrap(buf);
        outBuf = ByteBuffer.allocate(buf.length);
        processData(ib);

        // Pass the output buffer to document
        outBuf.flip();
        session.getDocument().processData(outBuf);
    }

    /**
     * Send raw data to the server.
     *
     * @param buf the data to send
     */
    public void sendRawData(byte[] buf) {
        session.sendRawData(buf);
    }

    /*===================*
     *  Telnet Commands  *
     *===================*/
    /**
     * process incoming data.
     *
     * @param ib input byte buffer
     */
    private void processData(ByteBuffer ib) {
        while (ib.hasRemaining()) {
            byte b = ib.get();

            processData(b);
        }
    }

    /**
     * Process byte according to current state
     *
     * @param b next byte
     */
    private void processData(byte b) {
        switch (state) {
        case INIT:
            if (TC_IAC == b) {
                log.debug("In IAC...");
                state = TelnetState.IAC;
            } else {
                outBuf.put(b);
            }
            break;
        case IAC:
            processIAC(b);
            break;
        case CMD:
            processCMD(b);
            break;
        case SUB:
            data.setSubOption(b);
            data.clearSubValue();
            state = TelnetState.SUB_OPT;
            break;
        case SUB_OPT:
            if (TC_IAC == b) {
                state = TelnetState.SUB_IAC;
            } else {
                data.putSubValue(b);
            }
            break;
        case SUB_IAC:
            if (TC_IAC == b) {
                data.putSubValue(b);
                state = TelnetState.SUB_OPT;
            } else if (TC_SE == b) {
                processSubOption();
                data.clearSubValue();
                state = TelnetState.INIT;
            }
            break;
        }
    }

    /**
     * Start processing telnet commands.
     *
     * @param b next byte
     */
    private void processIAC(byte b) {
        log.debug("Process IAC, b = " + b);
        switch (b) {
        case TC_IAC:
            outBuf.put(b);
            state = TelnetState.INIT;
            break;
        case TC_SB:
            state = TelnetState.SUB;
            break;
        default:
            data.setCmd(b);
            state = TelnetState.CMD;
            break;
        }
        log.debug("Process IAC, state = " + state);
    }

    /**
     * Process command data.
     *
     * @param b next byte
     */
    private void processCMD(byte b) {
        log.debug("Process CMD, b = " + b);
        switch (data.getCmd()) {
        case TC_WILL:
            processWILL(b);
            break;
        case TC_DO:
            processDO(b);
            break;
        }
        log.debug("Process CMD, state = " + state);
    }

    /**
     * Process WILL data.
     *
     * @param b next byte
     */
    private void processWILL(byte b) {
        log.debug("Process WILL, b = " + b);
        byte[] buf = { TC_IAC, TC_DONT, b };
        switch (b) {
        case TO_ECHO:
        case TO_SUPPRESS_GO_AHEAD:
            buf[1] = TC_DO;
            break;
        }
        sendRawData(buf);

        state = TelnetState.INIT;
        log.debug("Process WILL, state = " + state);
    }

    /**
     * Process DO data.
     *
     * @param b next byte
     */
    private void processDO(byte b) {
        log.debug("Process DO, b = " + b);
        byte[] buf = { TC_IAC, TC_WONT, b };
        switch (b) {
        case TO_BINARY:
        case TO_ECHO:
        case TO_SUPPRESS_GO_AHEAD:
            buf[1] = TC_WILL;
            break;
        }
        sendRawData(buf);

        state = TelnetState.INIT;
        log.debug("Process DO, state = " + state);
    }

    /**
     * Process sub option when all data are available.
     */
    private void processSubOption() {
        
        ByteBuffer value = data.getSubValue();
        value.flip();
        switch (data.getSubOption()) {
        case TO_TERMINAL_TYPE:
            if (value.get() == TERMINAL_TYPE_SEND) {
                byte[] head = { TC_IAC, TC_SB, TO_TERMINAL_TYPE, TERMINAL_TYPE_IS };
                byte[] ttype = "vt100".getBytes();
                byte[] tail = { TC_IAC, TC_SE };

                log.debug("Process sub option " + TERMINAL_TYPE_SEND);
                int size = head.length + ttype.length + tail.length;
                ByteBuffer bb = ByteBuffer.allocate(size);
                bb.put(head).put(ttype).put(tail);
                bb.flip();
                byte[] buf = new byte[size];
                bb.get(buf);

                sendRawData(buf);
            }
            break;
        }
    }
}

/**
 * TelnetData stores necessary telnet data,
 * like command, option and suboption.
 *
 * @author SuperMMX
 */
class TelnetData {
    private static final int SUB_VALUE_SIZE = 255;

    /**
     * Command
     */
    private byte cmd;
    /**
     * Option
     */
    private byte opt;
    /**
     * Sub option
     */
    private byte subopt;
    /**
     * The value of suboption
     */
    private ByteBuffer subvalue;

    public TelnetData() {
        subvalue = ByteBuffer.allocate(SUB_VALUE_SIZE);
    }

    public byte getCmd() {
        return cmd;
    }

    public void setCmd(byte cmd) {
        this.cmd = cmd;
    }

    public byte getOption() {
        return opt;
    }

    public void setOption(byte opt) {
        this.opt = opt;
    }

    public byte getSubOption() {
        return subopt;
    }

    public void setSubOption(byte subopt) {
        this.subopt = subopt;
    }

    public ByteBuffer getSubValue() {
        return subvalue;
    }

    /**
     * Append the byte to the suboption value
     *
     * @param b the next byte
     */
    public void putSubValue(byte b) {
        subvalue.put(b);
    }

    /**
     * Clear the value of suboption
     */
    public void clearSubValue() {
        subvalue.clear();
    }
}
