package org.supermmx.ginger.document;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory; 

import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.List;
import java.util.ArrayList;

/**
 * ANSI Control data, see ECMA-048 for more details.
 *
 * @author SuperMMX
 */
public class AnsiData {
    /**
     * Default buffer size for parameters and intermediate bytes
     */
    public static final int DEFAULT_SIZE = 256;

    /**
     * ESP
     */
    public static final byte ESC     = (byte) 27; // 0x1b
    /**
     * CSI of C0
     */
    public static final byte ESC_CSI = (byte) '['; // 05b

    /**
     * Start of the range of parameter value
     */
    public static final byte PARAM_START  = (byte) '0'; // 0x30
    /**
     * End of the range of parameter value
     */
    public static final byte PARAM_END    = (byte) '?'; // 0x3f

    /**
     * Start of the range of intermediate value
     */
    public static final byte INTER_START  = (byte) ' '; // 0x20
    /**
     * Start of the range of intermediate value
     */
    public static final byte INTER_END    = (byte) '/'; // 0x2f

    /**
     * Start of the final byte value
     */
    public static final byte FINAL_START  = (byte) '@'; // 0x40
    /**
     * End of the final byte value
     */
    public static final byte FINAL_END    = (byte) '~'; // 0x7e

    /**
     * Start of the independent control function value
     */
    public static final byte INDEPENDENT_START = (byte) '`'; // 0x60
    /**
     * End of the independent control function value
     */
    public static final byte INDEPENDENT_END   = (byte) '~'; // 0x7e

    private static Log log = LogFactory.getLog(AnsiData.class);

    /**
     * Parameter bytes, within the range [ '0', '?' ]
     */
    private ByteBuffer paramBuffer;
    /**
     * Intermediate bytes, within the range [ ' ', '/' ]
     */
    private ByteBuffer interBuffer;
    /**
     * The final byte, within the range [ '@', '~' ]
     */
    private byte finalByte;

    public AnsiData() {
        paramBuffer = ByteBuffer.allocate(DEFAULT_SIZE);
        interBuffer = ByteBuffer.allocate(DEFAULT_SIZE);

        finalByte = 0;
    }

    public ByteBuffer getParams() {
        return paramBuffer;
    }

    public void putParam(byte b) {
        paramBuffer.put(b);
    }

    public ByteBuffer getInters() {
        return interBuffer;
    }

    public void putInter(byte b) {
        interBuffer.put(b);
    }

    public byte getFinal() {
        return finalByte;
    }

    public void setFinal(byte b) {
        this.finalByte = b;
    }

    /**
     * Get numeric parameter values without default values.
     */
    public Integer[] getFreeNumericParams(int defaultValue) {
        paramBuffer.flip();

        List<Integer> res = new ArrayList<Integer>();
        int value = -1;
        while (paramBuffer.hasRemaining()) {
            byte b = paramBuffer.get();
            if (b == ';') {
                if (value != -1) {
                    res.add(value);
                } else {
                    res.add(defaultValue);
                }
                value = -1;
            } else if (b >= '0' && b <= '9') {
                if (value == -1) {
                    value = 0;
                }
                value *= 10;
                value += b - '0';
            }
        }
        if (value != -1) {
            res.add(value);
        }

        if (res.size() == 0) {
            res.add(defaultValue);
        }
        return res.toArray(new Integer[0]);
    }

    /**
     * Get numeric parameter values.
     *
     * @param values default values
     */
    public int[] getNumericParams(int ... values) {
        paramBuffer.flip();

        int[] res = null;
        int size = 1;

        if (values != null && values.length > 0) {
            size = values.length;
            res = Arrays.copyOf(values, size);
        } else {
            res = new int[size];
        }

        int index = 0;
        int value = -1;
        while (paramBuffer.hasRemaining()) {
            byte b = paramBuffer.get();
            if (b == ';') {
                if (value != -1) {
                    res[index] = value;
                }
                index ++;
                value = -1;
            } else if (b >= '0' && b <= '9') {
                if (value == -1) {
                    value = 0;
                }
                value *= 10;
                value += b - '0';
            }
        }
        if (value != -1) {
            res[index] = value;
        }

        return res;
    }

    /**
     * Get numeric value for the parameter.
     *
     * @param defaultValue default value to set
     *
     * @return the numeric value for parameter
     */
    public int getNumericParam(int defaultValue) {
        int[] res = getNumericParams(defaultValue);
        return res[0];
    }

    /**
     * Get string value for the parameter.
     *
     * @return the string value
     */
    public String getStringParam() {
        paramBuffer.flip();
        byte[] values = new byte[paramBuffer.remaining()];
        paramBuffer.get(values);

        return new String(values);
    }

    /**
     * Reset the ANSI data
     */
    public void reset() {
        finalByte = 0;

        paramBuffer.clear();
        interBuffer.clear();
    }
}
