package org.supermmx.ginger.document;

/**
 * The attribute of a segment.
 *
 * @author SuperMMX
 */
public class GingerAttribute {
    /**
     * Default attribute
     */
    public static final GingerAttribute DEFAULT_ATTRIBUTE =
        new GingerAttribute();

    /**
     * Reset
     */
    public static final int RESET = 0;
    /**
     * Bold
     */
    public static final int BOLD = 1;
    /**
     * Underscored
     */
    public static final int UNDERSCORED = 4;
    /**
     * Blinking
     */
    public static final int BLINKING = 5;
    /**
     * Reversed or negative
     */
    public static final int REVERSED = 7;
    /**
     * Normal intensity, not bold
     */
    public static final int NORMAL = 22;
    /**
     * Not underscored
     */
    public static final int NOT_UNDERSCORED = 24;
    /**
     * Not blinking
     */
    public static final int NOT_BLINKING = 25;
    /**
     * Not reversed or negative
     */
    public static final int NOT_REVERSED = 27;

    /**
     * Start value for foreground
     */
    public static final int FG_START = 30;
    /**
     * End value for foreground
     */
    public static final int FG_END = 39;
    /**
     * Start value for background
     */
    public static final int BG_START = 40;
    /**
     * End value for background
     */
    public static final int BG_END = 49;

    /**
     * Index for black
     */
    public static final int BLACK = 0;
    /**
     * Index for red
     */
    public static final int RED = 1;
    /**
     * Index for green
     */
    public static final int GREEN = 2;
    /**
     * Index for yellow
     */
    public static final int YELLOW = 3;
    /**
     * Index for blue
     */
    public static final int BLUE = 4;
    /**
     * Index for magenta
     */
    public static final int MAGENTA = 5;
    /**
     * Index for cyan
     */
    public static final int CYAN = 6;
    /**
     * Index for white
     */
    public static final int WHITE = 7;

    /**
     * Background color index
     */
    private int bg;
    /**
     * Foreground color index
     */
    private int fg;
    /**
     * Whether bold
     */
    private boolean bold;
    /**
     * Whether underscored
     */
    private boolean underscored;
    /**
     * Whether blinking
     */
    private boolean blinking;
    /**
     * Whether reversed image
     */
    private boolean reversed;

    public GingerAttribute() {
        reset();
    }

    public void reset() {
        bg = BLACK;
        fg = WHITE;

        bold = false;
        underscored = false;
        blinking = false;
        reversed = false;
    }

    public int getBackground() {
        return bg;
    }

    public GingerAttribute setBackground(int bg) {
        this.bg = bg;
        return this;
    }

    public int getForeground() {
        return fg;
    }

    public GingerAttribute setForeground(int fg) {
        this.fg = fg;
        return this;
    }

    public boolean isBold() {
        return bold;
    }

    public GingerAttribute setBold(boolean bold) {
        this.bold = bold;
        return this;
    }

    public boolean isUnderscored() {
        return underscored;
    }

    public GingerAttribute setUnderscored(boolean underscored) {
        this.underscored = underscored;
        return this;
    }

    public boolean isBlinking() {
        return blinking;
    }

    public GingerAttribute setBlinking(boolean blinking) {
        this.blinking = blinking;
        return this;
    }

    public boolean isReversed() {
        return reversed;
    }

    public GingerAttribute setReversed(boolean reversed) {
        this.reversed = reversed;
        return this;
    }

    public boolean equals(Object o) {
        if (!(o instanceof GingerAttribute)) {
            return false;
        }

        GingerAttribute attr = (GingerAttribute) o;
        if (fg != attr.fg) {
            return false;
        }
        if (bg != attr.bg) {
            return false;
        }
        if (bold != attr.bold) {
            return false;
        }
        if (underscored != attr.underscored) {
            return false;
        }
        if (blinking != attr.blinking) {
            return false;
        }
        if (reversed != attr.reversed) {
            return false;
        }
        return true;
    }

    public Object clone() {
        GingerAttribute attr = new GingerAttribute();
        attr.fg = fg;
        attr.bg = bg;
        attr.bold = bold;
        attr.blinking = blinking;
        attr.underscored = underscored;
        attr.reversed = reversed;

        return attr;
    }

    /**
     * Clone the current attribute.
     *
     * @return the new attribute instance
     */
    public GingerAttribute cloneAttribute() {
        return (GingerAttribute) clone();
    }

    /**
     * Set the corresponding value according to the ANSI definition.
     *
     * @param value the value defined in ANSI
     */
    public void setValue(int value) {
        switch (value) {
        case RESET:
            reset();
            break;
        case BOLD:
            setBold(true);
            break;
        case UNDERSCORED:
            setUnderscored(true);
            break;
        case BLINKING:
            setBlinking(true);
            break;
        case REVERSED:
            setReversed(true);
            break;
        case NORMAL:
            setBold(false);
            break;
        case NOT_UNDERSCORED:
            setUnderscored(false);
            break;
        case NOT_BLINKING:
            setBlinking(false);
            break;
        case NOT_REVERSED:
            setReversed(false);
            break;
        default:
            if (value >= FG_START && value <= FG_END) {
                setForeground(value - FG_START);
            } else if (value >= BG_START && value <= BG_END) {
                setBackground(value - BG_START);
            }
        }
    }

    /**
     * String representation of the attribute.
     *
     * @return string representation
     */
    public String toString() {
        StringBuilder sb = new StringBuilder();

        sb.append("Foreground: ").append(fg);
        sb.append(", Background: ").append(bg);
        sb.append(", Bold: ").append(bold);
        sb.append(", Underscored: ").append(underscored);
        sb.append(", Blinking: ").append(blinking);
        sb.append(", Reversed: ").append(reversed);

        return sb.toString();
    }

    /**
     * Create an attribute instance from string representation.
     *
     * @param s the string representation like "1;35;48"
     *
     * @return created attribute
     */
    public static GingerAttribute createAttribute(String s) {
        GingerAttribute attr = new GingerAttribute();

        String[] values = s.split(";");
        for (String svalue : values) {
            int value = 0;
            try {
                value = Integer.parseInt(svalue);
            } catch (Exception e) {
            }
            attr.setValue(value);
        }
        
        return attr;
    }

    /**
     * Create an attribute instance from a series of values
     *
     * @param values the array like [1, 35, 48]
     *
     * @return created attribute
     */
    public static GingerAttribute createAttribute(int ... values) {
        GingerAttribute attr = new GingerAttribute();
        for (int value : values) {
            attr.setValue(value);
        }
        return attr;
    }
}
