package org.supermmx.ginger.document;

import static org.supermmx.ginger.document.AnsiData.*;

import org.supermmx.ginger.GingerSession;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory; 

import java.util.Arrays;
import java.util.List;
import java.util.LinkedList;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.Lock;
import java.nio.ByteBuffer;
import java.io.OutputStream;
import java.io.FileOutputStream;
import java.io.BufferedOutputStream;

/**
 * The document that contains the texts and also the attribute.
 *
 * @author SuperMMX
 */
public class GingerDocument {
    /**
     * Default width
     */
    public static final int DEFAULT_WIDTH = 80;
    /**
     * Default height
     */
    public static final int DEFAULT_HEIGHT = 24;

    /**
     * Blank filler character
     */
    public static final byte BLANK_FILLER = (byte) ' ';

    /**
     * Character LF
     */
    public static final byte LINE_FEED        = (byte) '\n';
    /**
     * Character CR
     */
    public static final byte CARRIAGE_RETURN  = (byte) '\r';
    /**
     * Character BS
     */
    public static final byte BACKSPACE        = (byte) '\b';
    /**
     * Character BEL
     */
    public static final byte BELL             = (byte) 7;
    /**
     * Character HT
     */
    public static final byte TAB              = (byte) '\t';
    
    private static Log log = LogFactory.getLog(GingerDocument.class);

    /**
     * The associated session
     */
    private GingerSession session;

    /**
     * Current width
     */
    private int width;
    /**
     * Current height
     */
    private int height;

    /**
     * Current caret
     */
    private GingerCaret caret;

    /**
     * Attributes of the data
     */
    private List<GingerLine> lines;

    /**
     * ANSI state
     */
    private AnsiState ansiState;
    /**
     * ANSI data
     */
    private AnsiData ansiData;
    /**
     * Current attribute
     */
    private GingerAttribute attr;

    /**
     * Lock for data updating and rendering
     */
    private Lock lock;

    public GingerDocument() {
        this(DEFAULT_WIDTH, DEFAULT_HEIGHT, null);
    }

    public GingerDocument(int width, int height) {
        this(width, height, null);
    }

    public GingerDocument(GingerSession session) {
        this(DEFAULT_WIDTH, DEFAULT_HEIGHT, session);
    }

    public GingerDocument(int width, int height, GingerSession session) {
        this.session = session;

        this.width = width;
        this.height = height;

        lines = new LinkedList<GingerLine>();
        for (int i = 0; i < height; i ++) {
            lines.add(createLine());
        }
        
        caret = new GingerCaret(width, height);

        ansiState = AnsiState.INIT;
        ansiData = new AnsiData();

        attr = new GingerAttribute();

        lock = new ReentrantLock();
    }

    public GingerCaret getCaret() {
        return caret;
    }

    public int getWidth() {
        return width;
    }

    public int getHeight() {
        return height;
    }

    /**
     * Get the list of lines
     */
    public List<GingerLine> getLines() {
        return lines;
    }

    /**
     * Get the line of the specified row.
     *
     * @param row the row index
     *
     * @param the corresponding lone, null if not found
     */
    public GingerLine getLine(int row) {
        if (row < 0 || row >= lines.size()) {
            return null;
        }
        return lines.get(row);
    }

    /**
     * Current attribute
     */
    public GingerAttribute getAttribute() {
        return attr;
    }

    /**
     * Set current attribute
     */
    protected void setAttribute(GingerAttribute attr) {
        this.attr = attr;
    }

    /**
     * Obtain the data lock.
     */
    public void lock() {
        lock.lock();
    }

    /**
     * Release the data lock.
     */
    public void unlock() {
        lock.unlock();
    }

    /*=========================*
     *  Process incoming data  *
     *=========================*/
    /**
     * Process the incoming data.
     *
     * @author ib input byte buffer
     */
    public void processData(ByteBuffer ib) {
        lock();
        while (ib.hasRemaining()) {
            byte b = ib.get();

            processData(b);
        }
        unlock();

        // Render the document
        session.getView().repaint();

        try {
            OutputStream os = new BufferedOutputStream(new FileOutputStream("data.log"));
            for (int i = 0; i < height; i ++) {
                os.write(getLine(i).getData());
                os.write('\n');
            }
            os.flush();
            os.close();
        } catch (Exception e) {
            log.error("Writing data", e);
        }
    }

    /**
     * Process next value.
     *
     * @param b next value
     */
    private void processData(byte b) {
        switch (ansiState) {
        case INIT:
            processINIT(b);
            break;
        case ESC:
            if (AnsiData.ESC_CSI == b) {
                ansiState = AnsiState.CSI;
            } else if (b >= INDEPENDENT_START &&
                       b <= INDEPENDENT_END) {
                if (log.isDebugEnabled()) {
                    log.debug("Independent control function " + b);
                }
                ansiState = AnsiState.INIT;
            }
            break;
        case CSI:
            processCSI(b);
            break;
        case PARAMETER:
            processParameter(b);
            break;
        case INTERMEDIATE:
            processIntermediate(b);
            break;
        }
    }

    /**
     * Process init
     *
     * @param b next byte
     */
    private void processINIT(byte b) {
        switch (b) {
        case ESC:
            ansiState = AnsiState.ESC;
            break;
        case LINE_FEED:
            lineFeed();
            break;
        case CARRIAGE_RETURN:
            carriageReturn();
            break;
        case BACKSPACE:
            backspace();
            break;
        case BELL:
            bell();
            break;
        case TAB:
            tab();
            break;
        default:
            setValue(b);
        }
    }

    /**
     * Process CSI.
     *
     * @param b next value
     */
    private void processCSI(byte b) {
        if (b >= PARAM_START
            && b <= PARAM_END) {
            ansiState = AnsiState.PARAMETER;
            ansiData.putParam(b);
        } else if (b >= INTER_START
                   && b <= INTER_END) {
            ansiState = AnsiState.INTERMEDIATE;
            ansiData.putInter(b);
        } else if (b >= FINAL_START
                   && b <= FINAL_END) {
            ansiData.setFinal(b);
            parseAnsi();
            ansiState = AnsiState.INIT;
        }
    }

    /**
     * Process parameters
     *
     * @param b next value
     */
    private void processParameter(byte b) {
        if (b >= PARAM_START
            && b <= PARAM_END) {
            ansiData.putParam(b);
        } else if (b >= INTER_START
                   && b <= INTER_END) {
            ansiState = AnsiState.INTERMEDIATE;
            ansiData.putInter(b);
        } else if (b >= FINAL_START
                   && b <= FINAL_END) {
            ansiData.setFinal(b);
            parseAnsi();
            ansiState = AnsiState.INIT;
        }
    }

    /**
     * Process intermediate bytes
     *
     * @param b next value
     */
    private void processIntermediate(byte b) {
        if (b >= INTER_START
                   && b <= INTER_END) {
            ansiData.putInter(b);
        } else if (b >= FINAL_START
                   && b <= FINAL_END) {
            ansiData.setFinal(b);
            parseAnsi();
            ansiState = AnsiState.INIT;
        }
    }

    /**
     * Parse the whole ANSI data
     */
    private void parseAnsi() {
        ByteBuffer bb = ansiData.getParams().duplicate();
        bb.flip();
        byte[] ba = new byte[bb.remaining()];
        bb.get(ba);
        String s = new String(ba);

        if (log.isTraceEnabled()) {
            log.trace("ANSI Function " + (char) ansiData.getFinal()
                      + ", Param: " + s);
        }

        switch (ansiData.getFinal()) {
        case 'A':
            // Cursor Up
            caret.previousRow(ansiData.getNumericParam(1));
            break;
        case 'B':
            // Cursor Down
            caret.nextRow(ansiData.getNumericParam(1));
            break;
        case 'C':
            // Cursor Right
            caret.forwardColumn(ansiData.getNumericParam(1));
            break;
        case 'D':
            // Cursor Left
            caret.backwardColumn(ansiData.getNumericParam(1));
            break;
        case 'H':
            // Cursor Position
            int[] values = ansiData.getNumericParams(1, 1);
            caret.moveTo(values[1] - 1, values[0] - 1);
            break;
        case 'J':
            // Erase In Page
            eraseInScreen(ansiData.getNumericParam(0));
            break;
        case 'K':
            // Erase In Line
            eraseInLine(ansiData.getNumericParam(0));
            break;
        case 'L':
            // Insert Line
            insertLine(ansiData.getNumericParam(0));
            break;
        case 'M':
            // Delete Line
            deleteLine(ansiData.getNumericParam(0));
            break;
        case 'm':
            // Color
            for (int value : ansiData.getFreeNumericParams(0)) {
                attr.setValue(value);
            }

            if (log.isTraceEnabled()) {
                log.trace("Attribute = " + attr);
            }
            break;
        default:
            log.warn("ANSI function " + (char) ansiData.getFinal() + " not handled");
            break;
        }
        ansiData.reset();
    }

    /*=====================*
     *  Various Functions  *
     *=====================*/
    /**
     * Process line feed
     */
    private void lineFeed() {
        cursorDown(1);
    }

    /**
     * Move cursor n lines up. Scroll down if necessary.
     *
     * @param n n lines
     */
    private void cursorUp(int n) {
        int row = caret.getRow();
        if (row - n < 0) {
            // scroll down (n - row) lines
            int affectedRows = n - row;
            for (int i = 0; i < affectedRows; i ++) {
                lines.add(0, createLine());
                lines.remove(lines.size() - 1);
            }
        }
        caret.previousRow(n);
    }

    /**
     * Move cursor n lines down. Scroll up if necessary.
     *
     * @param n n lines
     */
    private void cursorDown(int n) {
        int row = caret.getRow();
        if (row + n >= height) {
            // scroll up (row + n - height + 1) lines
            int affectedRows = row + n - height + 1;
            for (int i = 0; i < affectedRows; i ++) {
                lines.add(createLine());
                lines.remove(0);
            }
        }
        caret.nextRow(n);
    }

    /**
     * Process carriage return
     */
    private void carriageReturn() {
        caret.moveToColumn(0);
    }

    /**
     * Process backspace
     */
    private void backspace() {
        caret.backwardColumn();
    }

    /**
     * Process bell
     */
    private void bell() {
    }

    /**
     * Process tab
     */
    private void tab() {
    }

    /**
     * Set value in current caret.
     *
     * @param b the value to set
     */
    private void setValue(byte b) {
        getLine(caret.getRow()).setValue(b);

        caret.forwardColumn();
    }

    /**
     * Clear screen.
     *
     * @param value 0, clear from current position to end of the page
     *              1, clear from page start to current position (inclusive)
     *              2, clear entire screen
     */
    private void eraseInScreen(int value) {
        int col = caret.getColumn();
        int row = caret.getRow();

        int start = 0;
        int end = 0;

        switch (value) {
        case 0:
            start = row + 1;
            end = height;
            break;
        case 1:
            start = 0;
            end = row;
            break;
        case 2:
            start = 0;
            end = height;
            // where to put the caret ?
            break;
        }

        if (value < 2) {
            getLine(row).erase(value);
        }
        for (int i = start; i < end; i ++) {
            lines.set(i, createLine());
        }
    }

    /**
     * Erase in line.
     *
     * @param value 0, clear from current position to the end of the line
     *              1, clear from the beginging of the line to current position
     *              2, clear the entire line
     */
    private void eraseInLine(int value) {
        int row = caret.getRow();

        GingerLine line = getLine(row);
        line.erase(value);
    }

    /**
     * Delete lines.
     *
     * @param n lines
     */
    private void deleteLine(int n) {
        int row = caret.getRow();
        if (n > height - row) {
            n = height - row;
        }
        // scroll up n lines
        for (int i = 0; i < n; i ++) {
            lines.remove(row);
        }
        for (int i = 0; i < n; i ++) {
            lines.add(createLine());
        }
        caret.moveToColumn(0);
    }

    /**
     * Create a line with empty data and attribute.
     *
     * @return the empty line
     */
    private GingerLine createLine() {
        return new GingerLine(this);
    }

    /**
     * Insert lines.
     *
     * @param n lines
     */
    private void insertLine(int n) {
        int row = caret.getRow();
        // scroll up n lines
        for (int i = 0; i < n; i ++) {
            lines.add(row, createLine());
            lines.remove(lines.size() - 1);
        }
        caret.moveToColumn(0);
    }

}
