package org.supermmx.ginger.document;

import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;

/**
 * A line of the document, which only contains the attibutes of
 * the segments.
 *
 * @author SuperMMX
 */
public class GingerLine {
    private GingerDocument doc;
    /**
     * Segments in this line
     */
    private List<GingerSegment> segments;
    /**
     * Line content
     */
    private byte[] data;

    /**
     * Construct the ginger line with the specified width.
     *
     * @param width the line width, which should be
                    the same as the document
     */
    public GingerLine(GingerDocument doc) {
        this(doc, (GingerSegment[]) null);
    }

    /**
     * For test
     */
    protected GingerLine(GingerDocument doc, GingerSegment ... segs) {
        this.doc = doc;

        segments = new ArrayList<GingerSegment>();
        if (segs == null || segs.length == 0) {
            segments.add(new GingerSegment(0, doc.getWidth(), doc.getAttribute()));
        } else {
            segments.addAll(Arrays.asList(segs));
        }

        data = new byte[doc.getWidth()];
        Arrays.fill(data, GingerDocument.BLANK_FILLER);
    }

    /**
     * Get the segments of this line.
     *
     * @return the list of the segments
     */
    public List<GingerSegment> getSegments() {
        return segments;
    }

    /**
     * Get the segment with the specified index.
     *
     * @param index the segment index
     *
     * @return the corresponding segment, null if not found
     */
    public GingerSegment getSegment(int index) {
        if (index < 0 || index >= segments.size()) {
            return null;
        }

        return segments.get(index);
    }

    /**
     * Get the underlying data.
     *
     * @return the data as byte array
     */
    public byte[] getData() {
        return data;
    }

    /**
     * Set the value in current column of the document within this line.
     * The current attribute of document is applied.
     *
     * @param b the value to set
     */
    public void setValue(byte b) {
        int col = doc.getCaret().getColumn();
        data[col] = b;

        int segIndex = getSegmentIndex(col);
        GingerSegment seg = getSegment(segIndex);
        GingerAttribute segAttr = seg.getAttribute();
        GingerAttribute docAttr = doc.getAttribute();
        if (docAttr.equals(segAttr)) {
            return ;
        }

        // merge the adjacent segments if necessary
        // or break the current segment if necessary.

        GingerSegment prevSeg = getSegment(segIndex - 1);
        GingerSegment nextSeg = getSegment(segIndex + 1);

        // Merge all three segments
        if (seg.getLength() == 1
            && prevSeg != null && prevSeg.getAttribute().equals(docAttr)
            && nextSeg != null && nextSeg.getAttribute().equals(docAttr)) {
            // prevSeg is kept
            prevSeg.setEnd(nextSeg.getEnd());

            // remove seg and nextSeg
            segments.remove(segIndex + 1);
            segments.remove(segIndex);

            return ;
        }

        // Merge with previous segment
        if (col == seg.getStart()
            && prevSeg != null
            && prevSeg.getAttribute().equals(docAttr)) {

            prevSeg.setEnd(col + 1);

            if (seg.getLength() == 1) {
                // if the length is 1, remove it
                segments.remove(segIndex);
            } else {
                seg.setStart(col + 1);
            }
            return ;
        }

        // Merge with next segment
        if (col == seg.getEnd() - 1
            && nextSeg != null
            && nextSeg.getAttribute().equals(docAttr)) {

            nextSeg.setStart(col);

            if (seg.getLength() == 1) {
                // if the length is 1, remove it
                segments.remove(segIndex);
            } else {
                seg.setEnd(col);
            }
            return ;
        }

        // No merging, update the attribute
        if (seg.getLength() == 1) {
            seg.setAttribute(docAttr);

            return ;
        }

        // create new segment
        GingerSegment newSeg = new GingerSegment(col, col + 1, docAttr);
        if (col == seg.getStart()) {
            // add before current
            segments.add(segIndex, newSeg);
            seg.setStart(col + 1);
        } else if (col == seg.getEnd() - 1) {
            // add after current
            segments.add(segIndex + 1, newSeg);
            seg.setEnd(col);
        } else {
            // break current one into three segments
            segments.add(segIndex + 1, newSeg);
            GingerSegment secondSeg = new GingerSegment(col + 1, seg.getEnd(),
                                                        seg.getAttribute());
            segments.add(segIndex + 2, secondSeg);

            seg.setEnd(col);
        }
    }

    /**
     * Get the segment index from the column.
     *
     * @param col current column
     *
     * @return the index of the segment that contains the column,
     *         -1 if not found, but should not happen
     */
    private int getSegmentIndex(int col) {
        int res = -1;
        int i = 0;
        for (GingerSegment seg : segments) {
            if (col >= seg.getStart()
                && col < seg.getStart() + seg.getLength()) {
                res = i;
                break;
            }
            i ++;
        }
        return res;
    }

    /**
     * Erase in line.
     *
     * @param value 0, clear from current position to the end of the line
     *              1, clear from the beginging of the line to current position
     *              2, clear the entire line
     */
    public void erase(int value) {
        int col = doc.getCaret().getColumn();
        int width = doc.getWidth();

        int segIndex = getSegmentIndex(col);
        GingerSegment seg = getSegment(segIndex);
        GingerAttribute segAttr = seg.getAttribute();
        GingerAttribute docAttr = doc.getAttribute();
        
        GingerSegment prevSeg = getSegment(segIndex - 1);
        GingerSegment nextSeg = getSegment(segIndex + 1);

        switch (value) {
        case 0:
            Arrays.fill(data, col, width, GingerDocument.BLANK_FILLER);

            // remove all the following segments
            for (int i = segments.size() - 1; i > segIndex; i --) {
                segments.remove(i);
            }

            if (col == seg.getStart()
                && prevSeg != null
                && docAttr.equals(prevSeg.getAttribute())) {
                // merge with previous one
                segments.remove(segIndex);
                prevSeg.setEnd(width);
            } else if (segAttr.equals(docAttr)
                       || col == seg.getStart()) {
                // current one is replaced
                seg.setEnd(width);
                seg.setAttribute(docAttr);
            } else {
                // create a new one
                seg.setEnd(col);
                GingerSegment newSeg = new GingerSegment(col, width, docAttr);
                segments.add(newSeg);
            }
            break;
        case 1:
            Arrays.fill(data, 0, col + 1, GingerDocument.BLANK_FILLER);

            // remove all the previous segments
            for (int i = segIndex - 1; i >= 0; i --) {
                segments.remove(i);
            }

            if (col == seg.getEnd() - 1
                && nextSeg != null
                && docAttr.equals(nextSeg.getAttribute())) {
                // merge with next one
                segments.remove(0);
                nextSeg.setStart(0);
            } else if (segAttr.equals(docAttr)
                       || col == seg.getEnd() - 1) {
                // current one is replaced
                seg.setStart(0);
                seg.setAttribute(docAttr);
            } else {
                // create a new one
                seg.setStart(col + 1);
                GingerSegment newSeg = new GingerSegment(0, col + 1, docAttr);
                segments.add(0, newSeg);
            }
            break;
        case 2:
            Arrays.fill(data, GingerDocument.BLANK_FILLER);
            segments.clear();
            segments.add(new GingerSegment(0, width, docAttr));
            break;
        }
    }

    public boolean equals(Object o) {
        if (!(o instanceof GingerLine)) {
            return false;
        }

        GingerLine line = (GingerLine) o;
        if (!Arrays.equals(data, line.data)) {
            return false;
        }

        return segments.equals(line.segments);
    }

    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("Data: ").append(Arrays.toString(data)).append(", Segments:\n");
        for (GingerSegment seg : segments) {
            sb.append("\t").append(seg).append("\n");
        }

        return sb.toString();
    }
}
