package org.supermmx.ginger.gui;

import org.supermmx.ginger.Ginger;
import org.supermmx.ginger.GingerSession;
import org.supermmx.ginger.GingerDestination;
import org.supermmx.ginger.view.GingerView;
import org.supermmx.ginger.gui.action.ShowQuickConnectAction;
import org.supermmx.ginger.gui.action.ExitAction;
import org.supermmx.ginger.gui.action.DisconnectAction;
import org.supermmx.ginger.gui.action.ConnectAction;
import org.supermmx.ginger.gui.action.CloseAction;
import org.supermmx.ginger.gui.action.ShowAboutAction;

import org.supermmx.peony.ui.PeonyMnemonicText;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory; 

import org.xnap.commons.i18n.I18nFactory;
import org.xnap.commons.i18n.I18n;

import java.util.List;
import java.util.ArrayList;
import java.awt.BorderLayout;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import javax.swing.JFrame;
import javax.swing.JTabbedPane;
import javax.swing.JMenuBar;
import javax.swing.JMenu;
import javax.swing.JMenuItem;

/**
 * Ginger main frame.
 *
 * @author SuperMMX
 */
public class GingerFrame extends JFrame {
    private static Log log = LogFactory.getLog(GingerFrame.class);

    private Ginger ginger;

    private JTabbedPane tabbedPane;
    private JMenuBar menubar;

    private I18n i18n;

    public GingerFrame(Ginger ginger) {
        super("Ginger");

        this.ginger = ginger;
        i18n = ginger.getI18n(getClass());

        createUI();
    }

    /**
     * Add a new session in the tabbed pane.
     * If the session of current tab is not connected,
     * it will be replaced with the new session.
     *
     * @param session the session to add
     */
    public void addSession(GingerSession session) {
        String title = session.getName();
        if (title == null || title.length() <= 0) {
            title = i18n.tr("<Empty>");
        }

        // get current tab
        int index = tabbedPane.getSelectedIndex();
        if (index >= 0) {
            GingerSession oldSession = getSession(index);
            if (!oldSession.isConnected()) {
                // replace with new session
                tabbedPane.setComponentAt(index, session.getView());
                tabbedPane.setTitleAt(index, title);

                return ;
            }
        }
        tabbedPane.add(title, session.getView());
        tabbedPane.setSelectedIndex(tabbedPane.getTabCount() - 1);
    }

    /**
     * Get all sessions.
     *
     * @return a list of all sessions
     */
    public List<GingerSession> getSessions() {
        List<GingerSession> sessions = new ArrayList<GingerSession>();
        for (int i = 0; i < tabbedPane.getTabCount(); i ++) {
            sessions.add(getSession(i));
        }

        return sessions;
    }

    /**
     * Get current selected session.
     *
     * @return current selected session
     */
    public GingerSession getSelectedSession() {
        return getSession(tabbedPane.getSelectedIndex());
    }

    /**
     * Get the session at the specified index.
     *
     * @param index the session index
     *
     * @return the session at the index
     */
    public GingerSession getSession(int index) {
        if (index < 0 || index >= tabbedPane.getTabCount()) {
            return null;
        }
        GingerView view = (GingerView)tabbedPane.getComponentAt(index);
        GingerSession session = view.getSession();

        return session;
    }

    /**
     * Close the current selected session.
     */
    public void closeSelectedSession() {
        closeSession(tabbedPane.getSelectedIndex());
    }

    /**
     * Close the session with the sepcified index.
     *
     * @param index the session index
     */
    public void closeSession(int index) {
        if (index < 0 || index >= tabbedPane.getTabCount()) {
            return ;
        }
        GingerView view = (GingerView)tabbedPane.getComponentAt(index);
        GingerSession session = view.getSession();

        session.disconnect();

        tabbedPane.remove(index);

        if (tabbedPane.getTabCount() == 0) {
            addSession(ginger.getEmptySession());
        }
    }

    private void createUI() {
        // menu
        createMenu();

        // tabbed pane
        tabbedPane = new JTabbedPane();
        
        getContentPane().add(tabbedPane);

        // misc
        addWindowListener(new WindowAdapter() {
                public void windowClosing(WindowEvent e) {
                    GingerFrame.this.ginger.exit();
                }
            });

        setSize(800, 600);
        setLocationRelativeTo(null);
        setVisible(true);
    }

    private void createMenu() {
        menubar = new JMenuBar();

        setJMenuBar(menubar);

        // connection menu
        JMenu connectionMenu = new JMenu();
        PeonyMnemonicText.applyMnemonicText(i18n.tr("&Session"),
                                            connectionMenu);
        menubar.add(connectionMenu);

        // quick connect
        connectionMenu.add(new ShowQuickConnectAction(ginger));
        // connect
        connectionMenu.add(new ConnectAction(ginger));
        // disconnect
        connectionMenu.add(new DisconnectAction(ginger));
        // close
        connectionMenu.add(new CloseAction(ginger));

        connectionMenu.addSeparator();

        // exit
        connectionMenu.add(new ExitAction(ginger));

        // help menu
        JMenu helpMenu = new JMenu();
        PeonyMnemonicText.applyMnemonicText(i18n.tr("&Help"),
                                            helpMenu);
        menubar.add(helpMenu);

        helpMenu.add(new ShowAboutAction(ginger));
    }

}
