package org.supermmx.ginger.view;

import org.supermmx.ginger.GingerSession;
import org.supermmx.ginger.document.GingerDocument;
import org.supermmx.ginger.document.GingerCaret;
import org.supermmx.ginger.document.GingerLine;
import org.supermmx.ginger.document.GingerSegment;
import org.supermmx.ginger.document.GingerAttribute;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory; 

import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.geom.Rectangle2D;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.KeyEvent;
import javax.swing.JPanel;
import javax.swing.RepaintManager;
import javax.swing.DebugGraphics;

/**
 * GingerView renders the ginger document.
 *
 * @author SuperMMX
 */
public class GingerView extends JPanel {
    public static Color DARK_RED = new Color(0x800000);
    public static Color DARK_GREEN = new Color(0x008000);
    public static Color DARK_YELLOW = new Color(0x808000);
    public static Color DARK_BLUE = new Color(0x000080);
    public static Color DARK_MAGENTA = new Color(0x800080);
    public static Color DARK_CYAN = new Color(0x008080);

    private static Log log = LogFactory.getLog(GingerView.class);

    private GingerSession session;

    private Color[] palette;

    public GingerView(GingerSession session) {
        this.session = session;

        palette = new Color[] {
            Color.BLACK, Color.GRAY,
            DARK_RED, Color.RED, 
            DARK_GREEN, Color.GREEN,
            DARK_YELLOW, Color.YELLOW,
            DARK_BLUE, Color.BLUE,
            DARK_MAGENTA, Color.MAGENTA,
            DARK_CYAN, Color.CYAN,
            Color.LIGHT_GRAY, Color.WHITE
        };

        createUI();
    }

    public GingerSession getSession() {
        return session;
    }

    private void createUI() {
        setFocusable(true);
        setRequestFocusEnabled(true);
        // Disable focus traversal keys like TAB/S-TAB etc.
        setFocusTraversalKeysEnabled(false);

        // JPanel will get the focus when mouse clicked
        addMouseListener(new MouseAdapter() {
                public void mousePressed(MouseEvent e) {
                    requestFocusInWindow();
                }
            });

        addKeyListener(new GingerViewKeyHandler());

        setFont(new Font("Simsun", Font.PLAIN, 16));

        /*
        // Debugging
        RepaintManager repaintManager = RepaintManager.currentManager(this);
        repaintManager.setDoubleBufferingEnabled(false);

        DebugGraphics.setFlashTime( 100 );
        DebugGraphics.setFlashCount( 2 );

        setDebugGraphicsOptions(DebugGraphics.BUFFERED_OPTION
                                | DebugGraphics.FLASH_OPTION
                                | DebugGraphics.LOG_OPTION);
        */
    }

    @Override
    public void paintComponent(Graphics g) {
        super.paintComponent(g);

        // to avoid weird background filling
        //g.setClip(0, 0, getWidth(), getHeight());
        g.setColor(Color.BLACK);
        g.fillRect(0, 0, getWidth(), getHeight());

        GingerDocument doc = session.getDocument();
        int width = doc.getWidth();
        int height = doc.getHeight();

        int x = 10;
        int y = 20;

        FontMetrics fm = g.getFontMetrics();
        int cw = fm.charWidth('M');
        int ch = fm.getHeight();

        // must be set to make sure everything is correct
        //g.setClip(0, 0, x + width * cw, y + height * ch - fm.getAscent());

        doc.lock();

        int index = 0;
        // Draw the document line by line
        for (GingerLine line : doc.getLines()) {
            index ++;

            drawLine(g, line, x, y);

            x = 10;
            y += ch;
        }

        x = 10;
        y = 20;

        g.setColor(Color.WHITE);

        // draw the caret
        GingerCaret caret = doc.getCaret();
        int cx = x + cw * caret.getColumn();
        // the position specified in drawString is the baseline.
        int cy = y + ch * caret.getRow() - fm.getAscent();
        g.drawLine(cx, cy, cx, cy + fm.getHeight());

        doc.unlock();
    }

    /**
     * Get foreground color based on bold.
     */
    private Color getColor(int index, GingerAttribute attr) {
        if (attr.isBold()) {
            index = 2 * index + 1;
        } else {
            index *= 2;
        }
        return palette[index];
    }

    private void drawLine(Graphics g, GingerLine line, int x, int y) {
        GingerDocument doc = session.getDocument();
        int width = doc.getWidth();
        int height = doc.getHeight();

        FontMetrics fm = g.getFontMetrics();
        int cw = fm.charWidth('M');
        int ch = fm.getHeight();

        // draw Chinese and ascii character separately
        byte[] data = line.getData();

        int start = 0;
        int end = 0;

        for (GingerSegment seg : line.getSegments()) {
            GingerAttribute attr = seg.getAttribute();

            start = seg.getStart();
            while (start < seg.getEnd() && data[start] != 0) {
                // find continuous Chinese or ascii
                boolean single = data[start] > 0;
                end = start + 1;
                while (end < seg.getEnd() && data[end] != 0) {
                    boolean isSingle = (data[end] > 0);
                    if ((single && !isSingle)
                        || (!single && isSingle
                            && ((end - start) % 2 == 0))) {
                        break;
                    }
                    end ++;
                }

                String charset = single ? "ISO-8859-1" : "GBK";
                String s = "";
                try {
                    s = new String(data, start, end - start, charset);
                } catch (Exception e) {
                    log.error("get string", e);
                    s = new String(data, start, end - start);
                }
                if (log.isTraceEnabled()) {
                    log.trace("String = \"" + s + "\"");
                }

                Rectangle2D rect = fm.getStringBounds(s, g);

                // fill with the background
                g.setColor(palette[attr.getBackground() * 2]);
                g.fillRect(x + (int)rect.getX(), y + (int)rect.getY(),
                           (int)rect.getWidth(), (int)rect.getHeight());

                // foreground text
                g.setColor(getColor(attr.getForeground(), attr));
                g.drawString(s, x, y);

                x += rect.getWidth();
                start = end;
            }
        }
    }
}
